/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.midi;

import jarre.midi.event.EventEnum;
import jarre.midi.event.EventMessage;
import jarre.midi.event.TempoChangeEvent;
import jarre.midi.event.TimeSignatureEvent;
import jarre.midi.sequencer.SequencerHelper;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.Patch;
import javax.sound.midi.Sequence;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Track;

public class MySequence extends Sequence {
    private Sequence sequence;
    private int minNote, maxNote;
    private boolean[] trackVisible;
    private boolean[] playBackground;
    private MyTrack[] tracks;
    private TempoChangeEvent[] tempoArray;
    private int[] timeMarkers;
    private MySequence(float divisionType, int resolution, Sequence sequence) throws InvalidMidiDataException {
	super(divisionType, resolution);
	this.sequence = sequence;
    }
    private MySequence(float divisionType, int resolution, int numTracks, Sequence sequence) throws InvalidMidiDataException {
	super(divisionType, resolution, numTracks);
	this.sequence = sequence;
    }
    private void computeMinMaxNote() {
	minNote = Integer.MAX_VALUE;
	maxNote = Integer.MIN_VALUE;
	for (Track track: sequence.getTracks()) {
	    int size = track.size();
	    for (int index = 0; index < size; index++) {
		MidiMessage message = track.get(index).getMessage();
		if (message instanceof ShortMessage) {
		    ShortMessage shortMessage = (ShortMessage)message;
		    switch (shortMessage.getCommand()) {
			case ShortMessage.NOTE_ON:
			case ShortMessage.NOTE_OFF:
			    int note = shortMessage.getData1();
			    if (minNote > note) minNote = note;
			    if (maxNote < note) maxNote = note;
		    }
		}
	    }
	}
    }
    private TempoChangeEvent[] createTempoArray() {
	ArrayList<TempoChangeEvent> tempoArray = new ArrayList<TempoChangeEvent>();
	for (MyTrack track: tracks)
	    tempoArray.addAll(track.getTempoList());
	Collections.sort(tempoArray, new Comparator<TempoChangeEvent>() {
	    @Override public int compare(TempoChangeEvent a, TempoChangeEvent b) {
		return a.tick - b.tick;
	    }
	});
	return tempoArray.toArray(new TempoChangeEvent[0]);
    }
    private int[] createTimeMarkers() {
	final long timeLength = getMicrosecondLength();
	int[] markers = new int[(int)(timeLength / 1000000L) + 1];
	int tick = 0;
	int dTime = 500000 / sequence.getResolution();
	int i = 0;
	long time = 0;
	while (time < timeLength) {
	    if (i < tempoArray.length) {
		TempoChangeEvent tempo = tempoArray[i];
		if (tempo.tick == tick) {
		    i++;
		    dTime = tempo.tempo / sequence.getResolution();
		}
	    }
	    int index = (int)(time / 1000000);
	    try {
		if (markers[index] == 0) markers[index] = tick;
	    } catch (Exception e) {
		break;
	    }
	    tick++;
	    time += dTime;
	}
	markers[0] = 0;
	return markers;
    }
    public static MySequence create(Sequence sequence) {
	MySequence seq = null;
	try {
	    seq = new MySequence(javax.sound.midi.Sequence.PPQ, 96, sequence);
	} catch (InvalidMidiDataException e) {
	    e.printStackTrace();
	}
	seq.computeMinMaxNote();
	Track[] tracks = sequence.getTracks();
	int length = tracks.length;
	seq.tracks = new MyTrack[length];
	seq.trackVisible = new boolean[length];
	seq.playBackground = new boolean[length];
	for (int i = 0; i < length; i++) {
	    seq.trackVisible[i] = true;
	    if (seq.hasNotes(i)) seq.playBackground[i] = true;
	    else seq.playBackground[i] = false;
	    seq.tracks[i] = new MyTrack(seq, tracks[i], i);
	}
	seq.tempoArray = seq.createTempoArray();
	seq.timeMarkers = seq.createTimeMarkers();
	return seq;
    }
    public static MySequence create(InputStream inStream) throws InvalidMidiDataException {
	try {
	    return create(MidiSystem.getSequence(inStream));
	} catch (IOException e) {
	}
	return null;
    }
    public void setPlayBackground(int trackNum, boolean play) {
	playBackground[trackNum] = play;
    }
    public boolean getPlayBackground(int trackNum) {
	return playBackground[trackNum];
    }
    public void setTrackVisible(int trackNum, boolean visible) {
	trackVisible[trackNum] = visible;
    }
    public boolean getTrackVisible(int trackNum) {
	return trackVisible[trackNum];
    }
    public int getMinNote() {
	return minNote;
    }
    public int getMaxNote() {
	return maxNote;
    }
    public boolean hasNotes(int numTrack) {
	Track track = sequence.getTracks()[numTrack];
	int size = track.size();
	for (int index = 0; index < size; index++) {
	    MidiMessage message = track.get(index).getMessage();
	    if (message instanceof ShortMessage) {
		ShortMessage msg = (ShortMessage)message;
		switch (msg.getCommand()) {
		    case ShortMessage.NOTE_ON:
		    case ShortMessage.NOTE_OFF:
			return true;
		}
	    }
	}
	return false;
    }
    public TimeSignatureEvent getTimeSignature() {
	for (int i = 0; i < tracks.length; i++) {
	    MyTrack singleTrack = tracks[i];
	    for (int k = 0; k < singleTrack.length; k++) {
		EventMessage[] events = singleTrack.getMessages(k);
		for (EventMessage message: events) {
		    if (message.type == EventEnum.TIME_SIGNATURE) {
			TimeSignatureEvent event = message.getEvent();
			return new TimeSignatureEvent(event.numerator, event.denumerator);
		    }
		}
	    }
	}
	return new TimeSignatureEvent(4, 4);
    }
    public SequencerHelper getSequenceHelper() {
	return new MyHelper();
    }
    public TempoChangeEvent[] getTempoArray() {
	return tempoArray;
    }
    public int[] getTimeMarkers() {
	return timeMarkers;
    }
    @Override public Track createTrack() {
	return sequence.createTrack();
    }
    @Override public boolean deleteTrack(Track track) {
	return sequence.deleteTrack(track);
    }
    @Override public float getDivisionType() {
	return sequence.getDivisionType();
    }
    @Override public long getMicrosecondLength() {
	return sequence.getMicrosecondLength();
    }
    @Override public Patch[] getPatchList() {
	return sequence.getPatchList();
    }
    @Override public int getResolution() {
	return sequence.getResolution();
    }
    @Override public long getTickLength() {
	return sequence.getTickLength();
    }
    @Override public Track[] getTracks() {
	return sequence.getTracks();
    }
    public MyTrack[] getMyTracks() {
	return tracks;
    }
    private class MyHelper extends SequencerHelper {
	protected MyHelper() {
	    super(MySequence.this);
	}
    }
}
