/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.midi;

import jarre.midi.event.EventEnum;
import jarre.midi.event.EventMessage;
import jarre.midi.event.NoteOffEvent;
import jarre.midi.event.NoteOnEvent;
import jarre.midi.event.ProgramChangeEvent;
import jarre.midi.event.TempoChangeEvent;
import jarre.midi.event.TimeSignatureEvent;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import javax.sound.midi.MetaMessage;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.SysexMessage;
import javax.sound.midi.Track;

public class MyTrack {
    public final int length;
    private static final EventMessage[] nullMessage = new EventMessage[] { new EventMessage(EventEnum.NONE) };
    private static final EventMessage[] exceptionMessage = new EventMessage[] { new EventMessage(EventEnum.ARRAY_INDEX_OUT_OF_BOUNDS) };
    private EventMessage messages[][];
    private LinkedList<EventMessage> messageQueue = new LinkedList<EventMessage>();
    private List<TempoChangeEvent> tempoList = new LinkedList<TempoChangeEvent>();
    private int currentTick;
    private HashMap<Integer, NoteOnEvent> noteOnMap = new HashMap<Integer, NoteOnEvent>();
    private ShortMessage previousProgramChange = null;
    private MetaMessage previousTempoChange = null;
    private MetaMessage previousTimeSignature = null;
    private final int trackNum;
    private final MySequence sequence;
    public MyTrack(MySequence sequence, Track track, int trackNum) {
	this.trackNum = trackNum;
	this.sequence = sequence;
	createArray(track);
	parseTrack(track);
	length = messages.length;
	System.gc();
    }
    public EventMessage[] getMessages(int timeTick) {
	EventMessage[] array;
	try {
	    array = messages[timeTick];
	} catch (ArrayIndexOutOfBoundsException e) {
	    return exceptionMessage;
	}
	if (array == null) return nullMessage;
	else return array;
    }
    public int getTrackNum() {
	return trackNum;
    }
    public List<TempoChangeEvent> getTempoList() {
	return Collections.unmodifiableList(tempoList);
    }
    private void createArray(Track track) {
	int size = track.size() - 1;
	int lastTick = (int)track.get(size).getTick();
	messages = new EventMessage[lastTick + 1][];

    }
    private void parseTrack(Track track) {
	MidiEvent event;
	MidiMessage message;
	long timeTick;
	int length = track.size();
	for (int i = 0; i < length; i++) {
	    event = track.get(i);
	    message = event.getMessage();
	    timeTick = event.getTick();
	    while (currentTick < timeTick) {
		copyMessages(currentTick);
		currentTick++;
	    }
	    if (message instanceof MetaMessage) messageQueue.add(parseMetaMessage(track, i));
	    else if (message instanceof SysexMessage) messageQueue.add(parseSysexMessage(track, i));
	    else if (message instanceof ShortMessage) {
		EventMessage eventMessage = parseShortMessage(track, i);
		if (eventMessage != null) messageQueue.add(eventMessage);
	    }
	}
	copyMessages(currentTick);
    }
    private void copyMessages(int index) {
	if (index >= 0 && ! messageQueue.isEmpty()) {
	    messages[index] = messageQueue.toArray(new EventMessage[0]);
	    messageQueue.clear();
	}
    }
    private EventMessage parseMetaMessage(Track track, int i) {
	MidiEvent midiEvent = track.get(i);
	MetaMessage message = (MetaMessage)midiEvent.getMessage();
	switch (message.getType()) {
	    case 0x51:
		TempoChangeEvent tempoChangeEvent = new TempoChangeEvent(message, previousTempoChange, trackNum, currentTick);
		previousTempoChange = message;
		tempoList.add(tempoChangeEvent);
		return new EventMessage(tempoChangeEvent);
	    case 0x58:
		TimeSignatureEvent timeSignatureEvent = new TimeSignatureEvent(message, previousTimeSignature, trackNum, currentTick);
		previousTimeSignature = message;
		return new EventMessage(timeSignatureEvent);
	    case 0x2F:
		return new EventMessage(EventEnum.END_OF_TRACK);
	    default:
		return new EventMessage(message);
	}
    }
    private EventMessage parseSysexMessage(Track track, int i) {
	SysexMessage message = (SysexMessage)track.get(i).getMessage();
	return new EventMessage(message);
    }
    private EventMessage parseShortMessage(Track track, int i) {
	ShortMessage message = (ShortMessage)track.get(i).getMessage();
	switch (message.getCommand()) {
	    case ShortMessage.NOTE_ON:
		int note = message.getData1();
		int velocity = message.getData2();
		if (velocity != 0) {
		    long length = track.get(i).getTick();
		    int k = i + 1;
		    for (; k < track.size(); k++) {
			MidiMessage midiMessage = track.get(k).getMessage();
			if (midiMessage instanceof ShortMessage) {
			    ShortMessage shortMessage = (ShortMessage)midiMessage;
			    if (shortMessage.getData1() == note) {
				if (shortMessage.getCommand() == ShortMessage.NOTE_OFF) {
				    break;
				}
				else if (shortMessage.getCommand() == ShortMessage.NOTE_ON && shortMessage.getData2() == 0) {
				    break;
				}
			    }
			}
		    }
		    if (k == track.size()) k--;
		    length = track.get(k).getTick() - length;
		    NoteOnEvent noteOnEvent = new NoteOnEvent(sequence, message, trackNum, currentTick, (int)length);
		    NoteOnEvent lastEvent = noteOnMap.remove(noteOnEvent.note);
		    if (lastEvent != null) messageQueue.add(new EventMessage(new NoteOffEvent(lastEvent)));
		    noteOnMap.put(noteOnEvent.note, noteOnEvent);
		    return new EventMessage(noteOnEvent);
		}
	    case ShortMessage.NOTE_OFF:
		NoteOnEvent noteOnEvent = noteOnMap.remove(message.getData1());
		if (noteOnEvent == null) return null;
		return new EventMessage(new NoteOffEvent(sequence, message, noteOnEvent, trackNum, currentTick));
	    case ShortMessage.PROGRAM_CHANGE:
		ProgramChangeEvent programChangeEvent = new ProgramChangeEvent(message, previousProgramChange, trackNum, currentTick);
		previousProgramChange = message;
		return new EventMessage(programChangeEvent);
	    default:
		return new EventMessage(message);
	}
    }
}
