/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.widgets;

import static jarre.i18n.I18n._;
import jarre.Jarre;
import jarre.midi.MySequence;
import jarre.widgets.common.ColorWidget;
import jarre.widgets.common.views.EItem;
import jarre.widgets.common.views.IControlProvider;
import jarre.widgets.common.views.IEnabledProvider;
import jarre.widgets.common.views.ILabelProvider;
import jarre.widgets.common.views.TableView;
import jarre.widgets.common.views.TableView.Item;
import jarre.widgets.common.views.TableView.SelectedEvent;
import jarre.widgets.piano.PianoAbstract;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.ColorDialog;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;

public class TracksWidget extends Composite {
    private Model[] model;
    private ColorMenu colorMenu;
    private TableView tableView;
    private MySequence sequence;
    private List<Listener> changeList = new LinkedList<Listener>();
    public TracksWidget(Composite parent, int style) {
	super(parent, style);
	setLayout(new FillLayout());
	colorMenu = new ColorMenu(getShell());
	tableView = new TableView(this, SWT.NONE, 4);
	tableView.setLabelProvider(new ILabelProvider() {
	    @Override public boolean getBoolValue(int row, int column) {
		Model m = model[row];
		if (m.trackNum == -1) return false;
		switch (column) {
		    case 0:
			return m.visible;
		    case 1:
			return m.playInBackground;
		    default:
			return false;
		}
	    }
	    @Override public String getText(int row, int column) {
		Model m = model[row];
		switch (column) {
		    case 2:
			if (m.trackNum == -1) return _("Input");
			else return _("Track") + " " + row;
		    default:
			return null;
		}
	    }
	    @Override public String getToolTip(int row, int column) {
		switch (column) {
		    case 0:
			return _("Show track.");
		    case 1:
			return _("Play track in background.");
		    default:
			return null;
		}
	    }	    
	});
	tableView.setControlProvider(new IControlProvider() {
	    @Override public Control createInstance(int row, int column) {
		Model m = model[row];
		if (m.trackNum == -1) return createInputItem(PianoAbstract.getTrackColor(-1));
		else return createColorWidget(m.color, m.trackNum);
	    }
	});
	tableView.setEnabledProvider(new IEnabledProvider() {
	    @Override public boolean getEnabled(int row, int column) {
		Model m = model[row];
		if (m.trackNum == -1 && (column == 0 || column == 1)) return false;
		return true;
	    }
	});
	tableView.addSelectedEvent(new SelectedEvent() {
	    @Override public void handleEvent(Item item, int row, int column) {
		Button button;
		Model m = model[row];
		switch (column) {
		    case 0:
			button = (Button)item.getControls()[column];
			m.visible = button.getSelection();
			if (m.visible) m.playInBackground = false;
			updateData(m);
			break;
		    case 1:
			button = (Button)item.getControls()[column];
			m.playInBackground = button.getSelection();
			if (m.playInBackground) m.visible = false;
			updateData(m);
			break;
		}
	    }
	    private void updateData(Model m) {
		sequence.setTrackVisible(m.trackNum, m.visible);
		sequence.setPlayBackground(m.trackNum, m.playInBackground);
		tableView.updateTable();
		sendChangeEvent(null);
	    }
	});
	tableView.setDefaultItems(EItem.CHECK_BUTTON, EItem.CHECK_BUTTON, EItem.LABEL, EItem.CONTROL);
	tableView.setFill(false, false, false, true);
	model = new Model[] { new Model(-1, null) };
	tableView.createItem(EItem.CHECK_BUTTON, EItem.CHECK_BUTTON, EItem.LABEL, EItem.CONTROL);
    }
    public void setSequence(MySequence sequence) {
	this.sequence = sequence;
	int numTracks = sequence.getTracks().length;
	ArrayList<Model> list = new ArrayList<Model>(numTracks);
	tableView.clear();
	int row = 0;
	for (int i = 0; i < numTracks; i++)
	    if (sequence.hasNotes(i)) {
		Color color = Jarre.colors[row++ % Jarre.colors.length];
		PianoAbstract.setTrackColor(color, row);
		list.add(new Model(i, color));
	    }
	list.add(new Model(-1, null));
	model = list.toArray(new Model[0]);
	tableView.createItems(model.length - 1);
	tableView.createItem(EItem.CHECK_BUTTON, EItem.CHECK_BUTTON, EItem.LABEL, EItem.CONTROL);
    }
    public void addTrackChangeListener(Listener listener) {
	synchronized (changeList) {
	    changeList.add(listener);
	}
    }
    private void sendChangeEvent(Event event) {
	Listener[] array;
	synchronized (changeList) {
	    array = changeList.toArray(new Listener[0]);
	}
	for (Listener e: array) e.handleEvent(event);
    }
    private Control createInputItem(Color color) {
	final ColorWidget colorWidget = new ColorWidget(tableView, SWT.NONE);
	colorWidget.setColor(color);
	colorWidget.addListener(SWT.MouseDown, new Listener() {
	    @Override public void handleEvent(Event event) {
		ColorDialog dialog = new ColorDialog(getShell(), SWT.NONE);
		dialog.setRGB(PianoAbstract.getTrackColor(-1).getRGB());
		RGB color = dialog.open();
		if (color != null) {
		    colorWidget.setColor(color.red, color.green, color.blue);
		    colorWidget.redraw();
		    colorWidget.update();
		    PianoAbstract.setTrackColor(new Color(getDisplay(), color), -1);
		}
	    } 
	});
	return colorWidget;
    }
    private ColorWidget createColorWidget(Color color, final int trackNum) {
	final ColorWidget colorWidget = new ColorWidget(tableView, SWT.NONE);
	colorWidget.setColor(color);
	colorWidget.addListener(SWT.MouseDown, new Listener() {
	    @Override public void handleEvent(Event event) {
		Listener listener = new Listener() {
		    @Override public void handleEvent(Event event) {
			Color color = colorMenu.getSelectedColor();
			colorWidget.setColor(color);
			PianoAbstract.setTrackColor(color, trackNum);
			colorWidget.redraw();
			colorMenu.removeListener(this);
			for (Listener listener: changeList) listener.handleEvent(event);
		    }
		};
		colorMenu.addListener(listener);
		Point pos = getDisplay().map(colorWidget, null, event.x, event.y);
		colorMenu.setLocation(pos.x, pos.y);
		colorMenu.open();
	    }
	});
	return colorWidget;
    }
    
    private class Model {
	public boolean visible = true;
	public boolean playInBackground = false;
	public Color color;
	public int trackNum;
	public Model(int trackNum, Color color) {
	    this.trackNum = trackNum;
	    this.color = color;
	}
    }
}
