/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.widgets.common;

import static jarre.i18n.I18n._;

import java.awt.SplashScreen;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StackLayout;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Shell;

public abstract class AppWindow {
    private Shell shell;
    private int returnCode = SWT.NONE;
    private Composite composite;
    private Composite area;
    private StackLayout areaLayout;
    private boolean fullScreen = false;
    private Animation animation;
    public AppWindow(Shell shell, String title) {
	this(shell, title, App.NONE);
    }
    public AppWindow(final Shell shell, String title, int flags) {
	this.shell = shell;
	shell.setText(title);
	shell.setLayout(new FillLayout());
	shell.addListener(SWT.Close, new Listener() {
	    @Override public void handleEvent(Event event) {
		animation.dispose();
		AppWindow.this.closeShell(event);
	    } 
	});
	if ((flags & App.CREATE_MENU_BAR) == App.CREATE_MENU_BAR) {
	    Menu bar = new Menu(shell, SWT.BAR);
	    createMenuBar(bar);
	    shell.setMenuBar(bar);
	}
	animation = new Animation(shell.getDisplay(), 48, 48, "data/process-working.png", shell.getDisplay().getSystemColor(SWT.COLOR_WIDGET_BACKGROUND));
	area = new Composite(shell, SWT.NONE);
	areaLayout = new StackLayout();
	area.setLayout(areaLayout);
	composite = new Composite(area, SWT.NONE);
	showOnTop(composite);
	createControl(composite);
    }
    public int open() {
	Display display = shell.getDisplay();
	shell.open();
	composite.layout();
	display.timerExec(1000, new Runnable() {
	    @Override public void run() {
		SplashScreen splash = SplashScreen.getSplashScreen();
		if (splash != null) splash.close();
	    }
	});
	while (!shell.isDisposed()) {
	    if (!display.readAndDispatch()) display.sleep();
	}
	return returnCode;
    }
    public void close(int returnCode) {
	this.returnCode = returnCode;
	shell.close();
    }
    public static void readAndDispatch(Shell shell) {
	while (! shell.isDisposed()) {
	    if (! shell.getDisplay().readAndDispatch()) break;
	}
    }
    public Shell getShell() {
	return shell;
    }
    public Display getDisplay() {
	return shell.getDisplay();
    }
    public void setFullscreen(boolean fullScreen) {
	this.fullScreen = fullScreen;
	shell.setFullScreen(fullScreen);
    }
    public boolean isFullscreen() {
	return fullScreen;
    }
    public void execTask(final Runnable runnable) {
	AppWindow.readAndDispatch(shell);
	final Composite composite = new Composite(area, SWT.NONE);
	composite.setLayout(new GridLayout());
	Image image = new Image(shell.getDisplay(), shell.getSize().x, shell.getSize().y);
	GC gc = new GC(shell);
	gc.copyArea(image, 0, 0);
	ImageData imageData = image.getImageData();
	PaletteData palette = imageData.palette;
	for (int x = 0; x < imageData.width; x++)
	    for (int y = 0; y < imageData.height; y++) {
		int pixel = imageData.getPixel(x, y);
		int r = ((pixel & palette.redMask) >>> 1) & palette.redMask;
		int g = ((pixel & palette.greenMask) >>> 1) & palette.greenMask;
	    	int b = ((pixel & palette.blueMask) >>> 1) & palette.blueMask;
	    	imageData.setPixel(x, y, r + g + b);
	    }
	final Image background = new Image(shell.getDisplay(), imageData);
	gc.dispose();
	image.dispose();
	composite.setBackgroundImage(background);
	Composite box = new Composite(composite, SWT.NONE);
	box.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, true));
	box.setLayout(new GridLayout());
	Label label = new Label(box, SWT.CENTER);
	label.setText(_("Please wait. Loading..."));
	final Label aniWidget = new Label(box, SWT.NONE);
	aniWidget.setLayoutData(new GridData(SWT.CENTER, SWT.CENTER, true, true));
	aniWidget.setImage(animation.getImage(0));
	getDisplay().timerExec(30, new Runnable() {
	    private final int length = animation.getSize();
	    private int pos = 1;
	    @Override public void run() {
		if (! aniWidget.isDisposed()) {
		    if ((pos % length) == 0) pos = 1;
		    aniWidget.setImage(animation.getImage(pos++));
		    aniWidget.redraw();
		    aniWidget.update();
		    getDisplay().timerExec(30, this);
		}
	    }
	});
	showOnTop(composite);
	new Thread() {
	    @Override public void run() {
		runnable.run();
		syncExec(new Runnable() {
		    @Override public void run() {
			aniWidget.dispose();
			showOnTop(AppWindow.this.composite);
			composite.dispose();
			background.dispose();
		    } 
		});
	    }
	}.start();
    }
    public void syncExec(Runnable runnable) {
	getDisplay().syncExec(runnable);
    }
    public void asyncExec(Runnable runnable) {
	getDisplay().asyncExec(runnable);
    }
    protected void createMenuBar(Menu menu) {
    }
    protected void closeShell(Event event) {
    }
    protected abstract void createControl(Composite parent);
    private void showOnTop(Composite composite) {
	areaLayout.topControl = composite;
	area.layout();
	shell.layout(true, true);
    }
}
