/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.widgets.common;

import jarre.reflect.Reflection;

import java.util.LinkedList;
import java.util.List;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;

public class PlayerWidget extends Composite {
    private Image startImage;
    private Image pauseImage;
    private Image skipBackwardImage;
    private Image seekBackwardImage;
    private Image seekForwardImage;
    private LinkedList<IWidget> playButtonList = new LinkedList<IWidget>();
    private boolean pause;
    private List<ActionListener> actionList = new LinkedList<ActionListener>();
    private String[] toolTips;
    private Listener playListener = new Listener(){
	@Override public void handleEvent(Event event) {
	    setPlay(! pause);
	    if (isPlay()) sendEvent(Action.PLAY);
	    else sendEvent(Action.PAUSE);
	}
    };
    private Listener skipBackwardListener = new Listener () {
	@Override public void handleEvent(Event event) {
	    sendEvent(Action.START);
	}
    };
    private Listener backwardListener = new Listener() {
	@Override public void handleEvent(Event event) {
	    sendEvent(Action.BACKWARD);
	}
    };
    private Listener forwardListener = new Listener() {
	@Override public void handleEvent(Event event) {
	    sendEvent(Action.FORWARD);
	}
    };
    private Listener focusInListener = new Listener() {
	@Override public void handleEvent(Event event) {
	    if (! SWT.getPlatform().equals("win32"))
		getShell().setFocus();
	}
    };
    public PlayerWidget(Composite parent, int style, String[] toolTips) {
	super(parent, style);
	startImage = new Image(getDisplay(), "data/media-playback-start.png");
	pauseImage = new Image(getDisplay(), "data/media-playback-pause.png");
	skipBackwardImage = new Image(getDisplay(), "data/media-skip-backward.png");
	seekBackwardImage = new Image(getDisplay(), "data/media-seek-backward.png");
	seekForwardImage = new Image(getDisplay(), "data/media-seek-forward.png");
	setLayout(new FillLayout());
	setToolTips(toolTips);
	createButtons(this, Button.class);
    }
    @Override public void dispose() {
	super.dispose();
	startImage.dispose();
	pauseImage.dispose();
	skipBackwardImage.dispose();
	seekBackwardImage.dispose();
	seekForwardImage.dispose();
    }
    public void setToolTips(String[] toolTips) {
	if (toolTips.length != 4) throw new RuntimeException("assert() error.");
	this.toolTips = toolTips;
    }
    public IWidget[] createButtons(Composite parent, Class<?> clazz) {
	IWidget[] array = new IWidget[4];
	try {
	    IWidget widget = Reflection.createInstance(IWidget.class, clazz);
	    array[0] = widget.constructor(parent, SWT.PUSH);
	    playButtonList.add(array[0]);
	    widget.setImage(startImage);
	    widget.addListener(SWT.Selection, playListener);
	    widget.addListener(SWT.FocusIn, focusInListener);
	    widget.setToolTipText(toolTips[0]);
	    array[1] = widget.constructor(parent, SWT.PUSH);
	    widget.setImage(skipBackwardImage);
	    widget.addListener(SWT.Selection, skipBackwardListener);
	    widget.addListener(SWT.FocusIn, focusInListener);
	    widget.setToolTipText(toolTips[1]);
	    array[2] = widget.constructor(parent, SWT.PUSH);
	    widget.setImage(seekBackwardImage);
	    widget.addListener(SWT.Selection, backwardListener);
	    widget.addListener(SWT.FocusIn, focusInListener);
	    widget.setToolTipText(toolTips[2]);
	    array[3] = widget.constructor(parent, SWT.PUSH);
	    widget.setImage(seekForwardImage);
	    widget.addListener(SWT.Selection, forwardListener);
	    widget.addListener(SWT.FocusIn, focusInListener);
	    widget.setToolTipText(toolTips[3]);
	    return array;
	} catch (NoSuchMethodException e) {
	    throw new RuntimeException(e);
	}
    }
    public boolean isPlay() {
	return pause;
    }
    public void setPlay(boolean p) {
	pause = p;
	for (IWidget w: playButtonList) {
	    if (w.isDisposed()) continue;
	    if (pause) w.setImage(pauseImage);
	    else w.setImage(startImage);
	}
    }
    public void addActionListener(ActionListener listener) {
	synchronized (actionList) {
	    actionList.add(listener);
	}
    }
    private void sendEvent(Action action) {
	LinkedList<ActionListener> temp;
	synchronized (actionList) {
	    temp = new LinkedList<ActionListener>(actionList);
	}
	for (ActionListener i: temp) i.handleEvent(action);
    }
    public enum Action { PLAY, PAUSE, START, BACKWARD, FORWARD };
    public static interface ActionListener {
	void handleEvent(Action action);
    }
    private static interface IWidget {
	IWidget constructor(Composite parent, int flags);
	boolean isDisposed();
	void setImage(Image image);
	void addListener(int type, Listener listener);
	void setToolTipText(String text);
    }
}
