/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.widgets.common.views;

import jarre.widgets.Utils;
import jarre.widgets.common.GridLayoutFactory;

import java.util.LinkedList;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;

public class TableView extends Composite {
    private Color bgColor = getDisplay().getSystemColor(SWT.COLOR_LIST_BACKGROUND);
    private LinkedList<Item> itemList = new LinkedList<Item>();
    private LinkedList<SelectedEvent> selectedEventList = new LinkedList<SelectedEvent>();
    private ILabelProvider labelProvider;
    private IControlProvider controlProvider;
    private IEnabledProvider enabledProvider;
    private int cols;
    private EItem[] defaultTypes;
    private boolean[] fillVertical;
    public TableView(Composite parent, int style, int cols) {
	super(parent, style);
	setBackground(bgColor);
	GridLayoutFactory.fillDefaults().columns(cols).hSpace(4).applyTo(this);
	this.cols = cols;
	fillVertical = new boolean[cols];
	for (int i = 0; i < cols; i++) fillVertical[i] = false;
    }
    public void clear() {
	for (Item item: itemList) item.dispose();
	itemList.clear();
    }
    public Item createItem(EItem ...itemTypes) {
	Item item = new Item(itemList.size(), itemTypes);
	itemList.add(item);
	return item;
    }
    public void createItems(int num, EItem ...itemTypes) {
	for (int i = 0; i < num; i++)
	    createItem(itemTypes);
    }
    public void createItems(int num) {
	createItems(num, defaultTypes);
    }
    public void setDefaultItems(EItem ...items) {
	if (items.length != cols) throw new RuntimeException("Wrong number of items.");
	defaultTypes = items;
    }
    public void setFill(boolean ...fillVertical) {
	if (fillVertical.length != cols) throw new RuntimeException("Wrong number of items.");
	this.fillVertical = fillVertical;
	for (Item item: itemList) item.setFill(fillVertical);
    }
    public void setLabelProvider(ILabelProvider labelProvider) {
	this.labelProvider = labelProvider;
	updateTable();
    }
    public void setControlProvider(IControlProvider controlProvider) {
	this.controlProvider = controlProvider;
    }
    public void setEnabledProvider(IEnabledProvider enabledProvider) {
	this.enabledProvider = enabledProvider;
	updateTable();
    }
    public void addSelectedEvent(SelectedEvent selectedEvent) {
	synchronized (selectedEventList) {
	    selectedEventList.add(selectedEvent);
	}
    }
    public void updateTable() {
	for (Item item: itemList) item.labelProviderChanged();
    }
    private void sendSelectedEvent(Item item, int column) {
	SelectedEvent[] array;
	synchronized (selectedEventList) {
	    array = selectedEventList.toArray(new SelectedEvent[0]);
	}
	for (SelectedEvent event: array) event.handleEvent(item, item.getRow(), column);
    }
    public class Item {
	protected Control[] controls;
	protected EItem[] items;
	protected int row;
	protected Item(int row, EItem ...items) {
	    if (items.length != cols) throw new RuntimeException("Wrong number of items.");
	    this.items = items;
	    this.row = row;
	    controls = new Control[items.length];
	    for (int i = 0; i < cols; i++) {
		Control widget = null;
		switch (items[i]) {
		    case NONE:
			widget = new Label(TableView.this, SWT.NONE);
			break;
		    case CONTROL:
			if (controlProvider != null) {
			    widget = controlProvider.createInstance(row, i);
			} else widget = new Label(TableView.this, SWT.NONE);
			labelProviderChanged(i, (Control)widget);
			break;
		    case LABEL:
			widget = new Label(TableView.this, SWT.NONE);
			labelProviderChanged(i, (Label)widget);
			break;
		    case CHECK_BUTTON:
			widget = new Button(TableView.this, SWT.CHECK);
			labelProviderChanged(i, (Button)widget);
			break;
		}
		widget.setBackground(bgColor);
		widget.setForeground(getDisplay().getSystemColor(SWT.COLOR_LIST_FOREGROUND));
		final int column = i;
		widget.addListener(SWT.MouseUp, new Listener() {
		    @Override public void handleEvent(Event event) {
			sendSelectedEvent(Item.this, column);
		    }
		});
		Utils.disableFocus(widget);
		controls[i] = widget;
	    }
	    setFill(fillVertical);
	}
	public void setFill(boolean ...fillVertical) {
	    if (items.length != cols) throw new RuntimeException("Wrong number of items.");
	    for (int i = 0; i < cols; i++)
		switch (items[i]) {
		    case LABEL:
			controls[i].setLayoutData(new GridData(SWT.FILL, SWT.CENTER, fillVertical[i], false));
			break;
		    default:
			controls[i].setLayoutData(new GridData(SWT.FILL, SWT.FILL, fillVertical[i], false));
		}
	}
	public void dispose() {
	    for (Control c: controls) c.dispose();
	}
	public EItem[] getItemTypes() {
	    return items;
	}
	public int getRow() {
	    return row;
	}
	public Control[] getControls() {
	    return controls;
	}
	protected void labelProviderChanged() {
	    if (labelProvider != null) {
		for (int i = 0; i < cols; i++)
		    switch (items[i]) {
		    	case CONTROL:
		    	    labelProviderChanged(i, (Control)controls[i]);
		    	    break;
		    	case LABEL:
		    	    labelProviderChanged(i, (Label)controls[i]);
		    	    break;
		    	case CHECK_BUTTON:
		    	    labelProviderChanged(i, (Button)controls[i]);
		    	    break;
		    }
	    }
	}
	private void labelProviderChanged(final int column, Control control) {
	    String text = labelProvider.getToolTip(row, column);
	    if (text != null) control.setToolTipText(text);
	    if (enabledProvider != null) control.setEnabled(enabledProvider.getEnabled(row, column));
	}
	private void labelProviderChanged(int column, Label label) {
	    String text = labelProvider.getText(row, column);
	    if (text != null) label.setText(text);
	    labelProviderChanged(column, (Control)label);
	}
	private void labelProviderChanged(int column, Button button) {
	    String text = labelProvider.getText(row, column);
	    if (text != null) button.setText(text);
	    button.setSelection(labelProvider.getBoolValue(row, column));
	    labelProviderChanged(column, (Control)button);
	}
    }
    public interface SelectedEvent {
	void handleEvent(Item item, int row, int column);
    }
}
