/*  Programmer:	Piotr Borek
 *  E-mail: piotrborek@op.pl
 *  Copyright 2009 Piotr Borek
 *
 *  Distributed under the terms of the GPL (GNU Public License)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jarre.widgets.piano;

import jarre.context.AbstractContext;
import jarre.context.IContext;
import jarre.midi.MySequence;
import jarre.midi.event.EventEnum;
import jarre.midi.event.EventMessage;
import jarre.midi.event.NoteOffEvent;
import jarre.midi.event.NoteOnEvent;
import jarre.midi.event.ProgramChangeEvent;
import jarre.midi.event.TempoChangeEvent;
import jarre.midi.event.TimeSignatureEvent;
import jarre.midi.sequencer.SeqTimer;
import jarre.midi.sequencer.SequencerAbstract;
import jarre.midi.sequencer.SequencerHelper;

import java.util.LinkedList;
import java.util.List;

import javax.sound.midi.MidiMessage;

import org.eclipse.swt.widgets.Display;

public class PianoSequencer extends SequencerAbstract {
    private volatile boolean playing = false;
    private volatile boolean waiting = false;
    private SequencerHelper helper;
    private final PianoKeyboard keyboard;
    private int lastTick = 0;
    private float fTick = 0;
    private int time = 0;
    private int metronomeTick = 0;
    private final SeqTimer sTimer = new SeqTimer(16) {
	private long pauseTime;
	@Override public void init() {
	    pauseTime = 0;
	}
	@Override public boolean process(int deltaTime) {
	    deltaTime -= pauseTime;
	    pauseTime = 0;
	    time += deltaTime;
	    fTick += dTimeToTick(deltaTime);
	    int tick = (int)fTick;
	    if (metronomeTick <= tick) {
		metronomeTick += sequence.getResolution();
		if (metronome) dispatchMetronomeMessage();
	    }
	    if (tick != lastTick) {
		List<EventMessage> messageList = null;
		for (int i = lastTick; i < tick; i++) {
		    helper.setCurrentTick(i);
		    List<EventMessage> messages = helper.getCurrentMessages();
		    if (messages != null) {
			if (messageList == null) messageList = new LinkedList<EventMessage>();
			messageList.addAll(messages);
		    }
		}
		if (messageList != null) {
		    for (EventMessage message: messageList)
			processMessage(message);
		    Display.getDefault().asyncExec(drawEvent);
		    if (followMode) {
			pauseTime = System.currentTimeMillis();
			followYouMode(messageList, lastTick);
			pauseTime = System.currentTimeMillis() - pauseTime;
		    }
		}
		else {
		    playing = false;
		    Display.getDefault().asyncExec(EOTEvent);
		}
		lastTick = tick;
	    }
	    return playing;
	}
	private void followYouMode(List<EventMessage> messageList, int lastTick) {
	    for (EventMessage message: helper.getActiveMessages())
		if (message.type == EventEnum.NOTE_ON) {
		    NoteOnEvent noteOn = message.getEvent();
		    if (noteOn.tick < lastTick)
			sendNoteOffEvent(new NoteOffEvent(noteOn));
		}
	    Display.getDefault().asyncExec(drawEvent);
	    synchronized (sTimer) {
		try {
		    waiting = true;
		    while (followMode && isNotFullMatch(messageList))
			wait();
		} catch (InterruptedException e) {
		} finally {
		    waiting = false;
		}
	    }
	    if (followMode) {
		for (EventMessage message: messageList)
		    if (message.type == EventEnum.NOTE_ON)
			sendNoteOffEvent(new NoteOffEvent(message.<NoteOnEvent>getEvent()));
	    } else {
		for (EventMessage message: helper.getActiveMessages())
		    if (message.type == EventEnum.NOTE_ON)
			sendNoteOnEvent(message.<NoteOnEvent>getEvent());
	    }
	}
	private boolean isNotFullMatch(List<EventMessage> messageList) {
	    boolean notFullMatch = false;
	    for (EventMessage message: messageList)
		if (message.type == EventEnum.NOTE_ON) {
		    NoteOnEvent noteOn = message.getEvent();
		    if (noteOn.isVisible())
			if (! keyboard.isRawNoteOn(noteOn.note))
			    notFullMatch = true;
		}
	    return notFullMatch;
	}
	private Runnable drawEvent = new Runnable() {
	    @Override public void run() {
		sendDrawEvent(helper.getCurrentTick());
	    }
	};
	private Runnable EOTEvent = new Runnable() {
	    @Override public void run() {
		sendEndOfTracksEvent();
		sendAllNotesOffEvent();
	    }
	};
    };
    public PianoSequencer(PianoKeyboard keyboard) {
	this.keyboard = keyboard;
    }
    public int getTime() {
	return time;
    }
    @Override public void setSequence(MySequence sequence) {
	super.setSequence(sequence);
	helper = sequence.getSequenceHelper();
	helper.setCurrentTick(0);
	lastTick = 0;
	fTick = 0;
	time = 0;
	metronomeTick = 0;
    }
    @Override public int getCurrentPos() {
	if (helper == null) return 0;
	else return helper.getCurrentTick();
    }
    @Override public void goBackward(int ticks) {
	playerContext.enter();
	time += dTicksToTime(-ticks);
	List<EventMessage> list = helper.goBackward(ticks);
	for (EventMessage message: list) processMessage(message, false);
	playerContext.exit();
    }
    @Override public void goForward(int ticks) {
	playerContext.enter();
	time += dTicksToTime(ticks);
	List<EventMessage> list = helper.goForward(ticks);
	for (EventMessage message: list) processMessage(message);
	playerContext.exit();
    }
    @Override public void goToPosition(int seconds) {
	if (sequence == null) return;
	playerContext.enter();
	time = seconds * 1000;
	int tick = helper.getCurrentTick() - sequence.getTimeMarkers()[seconds];
	if (tick > 0) {
	    List<EventMessage> list = helper.goBackward(tick);
	    for (EventMessage message: list) processMessage(message, false);
	}
	else {
	    List<EventMessage> list = helper.goForward(-tick);
	    for (EventMessage message: list) processMessage(message);
	}
	playerContext.exit();
    }
    @Override public boolean isPlaying() {
	return playing;
    }
    @Override public boolean isWaiting() {
	return waiting;
    }
    @Override public void noteChangeSignal() {
	synchronized (sTimer) {
	    sTimer.notify();
	}
    }
    @Override public void setFollowMode(boolean followMode) {
	super.setFollowMode(followMode);
	noteChangeSignal();
    }
    @Override public void pause() {
	boolean tmp = followMode;
	playing = false;
	try {
	    if (followMode) {
		followMode = false;
		noteChangeSignal();
	    }
	    while (sTimer.isRunning())
		Display.getCurrent().readAndDispatch();
	    sTimer.join();
	} catch (InterruptedException e) {
	} finally {
	    allSoundOff();
	    followMode = tmp;
	}
    }
    @Override public void play() {
	if (sequence != null) {
	    playing = true;
	    sTimer.start();
	}
    }
    private int dTicksToTime(int deltaTick) {
	final int tempo = this.tempo;
	final int currentTick = helper.getCurrentTick();
	int tick = currentTick;
	int time = 0;
	final int newPos = tick + deltaTick;
	if (deltaTick > 0) {
	    for (;tick <= newPos; tick++) {
		helper.setCurrentTick(tick);
		TempoChangeEvent tempoEvent = getTempoChange();
		if (tempoEvent != null) setTempo(tempoEvent.tempo);
		time += this.tempo / 1000;
	    }
	}
	else {
	    if (newPos < 0) return -this.time;
	    for (;tick >= newPos; tick--) {
		time -= this.tempo / 1000;
		helper.setCurrentTick(tick);
		TempoChangeEvent tempoEvent = getTempoChange();
		if (tempoEvent != null) setTempo(tempoEvent.previousTempo);
	    }
	}
	this.tempo = tempo;
	helper.setCurrentTick(currentTick);
	return time;
    }
    private float dTimeToTick(int deltaTime) {
	return (10.0f * getSpeed() * deltaTime) / this.tempo;
    }
    private TempoChangeEvent getTempoChange() {
	List<EventMessage> messageList = helper.getCurrentMessages();
	if (messageList == null) return null;
	for (EventMessage message: messageList)
	    switch (message.type) {
		case TEMPO_CHANGE:
		    return message.getEvent();
	    }
	return null;
    }
    private void processMessage(EventMessage message) {
	processMessage(message, true);
    }
    private void processMessage(EventMessage message, boolean goForward) {
	switch (message.type) {
	    case NOTE_ON:
		NoteOnEvent noteOn = message.getEvent();
		sendNoteOnEvent(noteOn);
		if (playerContext.getValue()) dispatchMessage(noteOn.midiMessage, -1);
		break;
	    case NOTE_OFF:
		NoteOffEvent noteOff = message.getEvent();
		sendNoteOffEvent(noteOff);
		dispatchMessage(noteOff.midiMessage, -1);
		break;
	    case OTHER_SHORT_MESSAGE:
		dispatchMessage(message.<MidiMessage>getEvent(), -1);
		break;
	    case TEMPO_CHANGE:
		if (goForward)
		    setTempo(message.<TempoChangeEvent>getEvent().tempo);
		else
		    setTempo(message.<TempoChangeEvent>getEvent().previousTempo);
		break;
	    case TIME_SIGNATURE:
		if (goForward)
		    setTimeSignature(message.<TimeSignatureEvent>getEvent());
		else {
		    TimeSignatureEvent event = message.getEvent();
		    setTimeSignature(event.previousNumerator, event.previousDenumerator);
		}
		break;
	    case PROGRAM_CHANGE:
		if (! ignoreProgramChange) {
		    if (goForward)
			dispatchMessage(message.<ProgramChangeEvent>getEvent().midiMessage, -1);
		    else
			dispatchMessage(message.<ProgramChangeEvent>getEvent().previousMidiMessage, -1);
		}
		break;
	}
    }
    private IContext<Boolean> playerContext = new AbstractContext<Boolean>() {
	@Override public void enter() {
	    if (playing) {
		pause();
		allSoundOff();
		pushContext(true);
	    } else pushContext(false);
	    sendAllNotesOffEvent();
	}
	@Override public void exit() {
	    fTick = helper.getCurrentTick();
	    lastTick = (int)fTick;
	    if (lastTick > sequence.getTickLength()) {
		lastTick = (int)sequence.getTickLength();
		fTick = lastTick;
		time = (int)(sequence.getMicrosecondLength() / 1000);
		helper.setCurrentTick(lastTick);
	    }
	    metronomeTick = lastTick - 1;
	    sendMoveEvent(helper.getCurrentTick());
	    Boolean p = popContext();
	    if (p == null) return;
	    if (p) play();
	}
	@Override public Boolean getValue() {
	    Boolean p = super.getValue();
	    if (p == null) return playing;
	    return p;
	}
    };
}
